#!/usr/bin/python
###-----------------------------------------------------------------------------###
#
#    Project:TCRead.py
#    File: TCRead.py    
#    Written By PiPlay
#    Licensed under Creative Commons Attribute CC-BY
#    V0.4  - March 2020
#    November 2019 - Mod to Curses setip to deal with windows smaller than display area.
#       March 2020 - Added delay after Mux setup and chip setup before aquisition.
#
###-----------------------------------------------------------------------------###

import spidev
import time
import os
import struct
import RPi.GPIO as GPIO
import psutil
import datetime
import configparser
import time
import curses

from collections import namedtuple

from TC import *
from TCspi import *
from TCtempconvert import *


config = configparser.ConfigParser()
config.sections()

config.read('TCRead.ini')

ChannelStruct = namedtuple("ChannelStruct", "status, thermocouple_type, name")
lclconf = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]

###-----------------------------------------------------------------------------###

def check_channel_on( chanstatus ):

    chanstatus = chanstatus.upper()    
    if ( chanstatus == "ON" ):
        return 1
    else:
        return 0

###-----------------------------------------------------------------------------###

def get_thermocoupletype(channel):
    global lclconf

    # Note First in list below is text string passed from ini file
    # Second in list below is corresponding setting for chip from TC.py header
    switcher = {
            "TYPE_B": TYPE_B,
            "TYPE_E": TYPE_E,
            "TYPE_J": TYPE_J,
            "TYPE_K": TYPE_K,
            "TYPE_N": TYPE_N,
            "TYPE_R": TYPE_R,
            "TYPE_S": TYPE_S,
            "TYPE_T": TYPE_T
        }

    tctype = switcher.get( lclconf[channel].thermocouple_type )
    return tctype

###-----------------------------------------------------------------------------###

def get_padding( instring ):
    mxlength = 20

    outstring = ""
    thislength = len( instring ) 
    for loop in range( 0, mxlength - thislength - 1 ):
        outstring = outstring + " "
    return outstring

###-----------------------------------------------------------------------------###

def store_setup( curchan ):
    global lclconf

    thischannel = "Channel" + str(curchan)
    if thischannel in config:
        mystatus = config[thischannel]['status']
        mytctype = config[thischannel]['thermocouple_type']
        myname = config[thischannel]['name']
        lclconf[curchan] = ChannelStruct( mystatus, mytctype, myname )
    else:
        mystatus = "OFF"
        mytctype = "NONE"
        myname = "Disabled"
        lclconf[curchan] = ChannelStruct( mystatus, mytctype, myname )


###-----------------------------------------------------------------------------###

def output_to_screen( window, arraytemp, nextacq, nextlog ):
    global lclconf

    now = datetime.datetime.now()
    mytime = now.strftime("%Y-%m-%d %H:%M")

    window.addstr( 0,0, "*****************************************************************" )
    window.addstr( 1,0, "*    TCRead.py V0.3 - 16 channel thermocouple board." )
    window.addstr( 2,0, "*    %s" % (mytime) )
    window.addstr( 3,0, "*       Time until next display    :%s             " % ( nextacq ) )
    window.addstr( 4,0, "*    Time until next logging    :%s                  " % ( nextlog ) )
    window.addstr( 5,0, "*****************************************************************" )
    
    displayline = 7
    for channel in range(0,16):
        if ( check_channel_on( lclconf[channel].status ) ):
            textspacing = 20
            currenttemp = arraytemp[channel]
            padding = get_padding( lclconf[channel].name )
            if currenttemp != ERRTEMP:
                string1 = "%2d ( " % ( channel )
                string2 = " )%s - Temperature: %0.2f deg C    " % ( padding, currenttemp )
                outstring = string1 + lclconf[channel].name + string2 
                window.addstr( displayline,0, outstring )
                displayline += 1
            else:
                string1 = "%2d ( " % ( channel )
                string2 = " )%s - Temperature: FAULT     " % ( padding ) 
                outstring = string1 + lclconf[channel].name + string2 
                window.addstr( displayline,0, outstring )
                displayline += 1
            
    currenttemp = arraytemp[16]
    window.addstr( displayline+1,0, "Cold Junction Temperature is:\t\t %0.2f deg C" % ( currenttemp ) )
    window.addstr( displayline+4,0, "Hit X or x to Terminate Program :" )
    window.refresh()

###-----------------------------------------------------------------------------###

def output_to_log( window, arraytemp, logfile, logfile_maximum, currentepochtime ):

    outstring = ""
    logfile_maximum = float(logfile_maximum)*1024*1024

    if ( os.path.exists( logfile ) ):    
        # If Logfile is greater than maximum roll log 
        filesize = os.stat(logfile).st_size 
        if( filesize > logfile_maximum ):
            timestring = time.strftime("%Y%m%d%H%M%S", time.localtime(currentepochtime) )
            shortlogfile = logfile.strip('.csv')
            outputfile =  (".%s%s.csv" % (shortlogfile, timestring ) )
            os.rename( logfile, outputfile )      

    if ( os.path.exists( logfile ) ):    
        fh = open( logfile, "a" )
    else:
        fh = open( logfile, "w" )
        outstring = "Date/Time,"
        for channel in range(0,16):
            if ( check_channel_on( lclconf[channel].status ) ):
                outstring  += lclconf[channel].name
                if ( channel != 16 ):
                    outstring += ','
        outstring += "Cold Junction"    
        fh.write( outstring )    
        fh.write( '\n' )

    timestring = time.strftime("%d/%m/%Y %H:%M:%S", time.localtime(currentepochtime) )
    outstring = timestring
    outstring += ','

    for channel in range(0,16):
        if ( check_channel_on( lclconf[channel].status ) ):
            currenttemp = arraytemp[channel]
            if currenttemp != ERRTEMP:
                outstring += str(currenttemp)
                outstring += ","
            else:
                outstring += "FAULT"
                outstring += ","
    outstring += str(arraytemp[16])
    fh.write( outstring )
    fh.write( '\n' )

    fh.close()

###-----------------------------------------------------------------------------###

def set_mux_address( address ):
    address = address & 0b00001111
    value = address & ( 1 << 3 )
    
    if ( value ):
        GPIO.output(MUX1_EN, GPIO.LOW)
        GPIO.output(MUX2_EN, GPIO.HIGH)
    else:
        GPIO.output(MUX1_EN, GPIO.HIGH)
        GPIO.output(MUX2_EN, GPIO.LOW)

    value = address & ( 1 << 2 )
    if ( value ):
        GPIO.output(MUX_A2, GPIO.HIGH) 
    else:
        GPIO.output(MUX_A2, GPIO.LOW) 
        
    value = address & ( 1 << 1 )
    if ( value ):
        GPIO.output(MUX_A1, GPIO.HIGH) 
    else:
        GPIO.output(MUX_A1, GPIO.LOW) 

    value = address & ( 1 << 0 )
    if ( value ):
        GPIO.output(MUX_A0, GPIO.HIGH) 
    else:
        GPIO.output(MUX_A0, GPIO.LOW) 


###-----------------------------------------------------------------------------###

def setupcomms():
    try:
        spi = spidev.SpiDev()
        spi.open(0,0)
        spi.max_speed_hz = 7629
        spi.mode = 1 

        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BOARD)

        GPIO.setup(FAULT, GPIO.IN )
        GPIO.setup(READY, GPIO.IN )
        GPIO.setup( MUX1_EN, GPIO.OUT ) 
        GPIO.setup( MUX2_EN, GPIO.OUT ) 
        GPIO.setup( MUX_A0, GPIO.OUT ) 
        GPIO.setup( MUX_A1, GPIO.OUT ) 
        GPIO.setup( MUX_A2, GPIO.OUT ) 

    except Exception as e:
        print( "Error occurred : " + str(e) )
        exit

    writereg( spi, MASK, OPNFLTMSK )
    
    return spi

###-----------------------------------------------------------------------------###

def acquire_temperatures(spi):
    global lclconf

    temp = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
    for channel in range(0,17):
        time.sleep(0.1)    
        if channel == 16:
            cjresult = get_cjtemperature(spi)
            precjtemp = cjconverttocelsius( cjresult[1], cjresult[2] )
            temp[channel] = round( precjtemp, 2 )
        else:
            if ( check_channel_on( lclconf[channel].status ) ):
                current_tctype = get_thermocoupletype(channel)
                set_mux_address( channel )
                writereg( spi, CR1, SAMP_4 | current_tctype ) 
                writereg( spi, CR0, ONESHOT | OCFAULT_ENLT5 | NR50HZ ) 
                time.sleep(0.5)

                waitcount = 0
                outstat = 0
                
                while ( GPIO.input(READY) == 1 & outstat == 0 ):
                    waitcount = waitcount + 1
                    if ( waitcount > 100 ):
                        print( "\r   No Response from Main Chip", outstat )
                        time.sleep(5)
                    outstat = 1
                    time.sleep(0.1)    
                
                result = get_temperature(spi)
                pretemp = converttocelsius( result[1], result[2], result[3] )

                if ( GPIO.input(FAULT) ):
                    temp[channel] = round( pretemp, 2 )
                else:
                    temp[channel] = ERRTEMP
    return temp


###-----------------------------------------------------------------------------###
# Wrapped function for Curses library so it will automatically clean up on exit.
# Main work is done here.

def mainlinecode( window ):

        # Do system setups and gather configuration data
        spi = setupcomms()
        arraytemp = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        
        window.clear()
        curses.resizeterm(70, 80)
        window.refresh()
        window.nodelay(True)

        for thischan in range( 0, 16 ):
            store_setup( thischan )

        display_interval = float(config['Default']['display_interval'])
        logging_interval = float(config['Default']['logging_interval'])
        logoutput = config['Default']['logoutput']
        logoutput = logoutput.upper()
        logfile_maximum = config['Default']['logfile_maximum'] 
        logfile = config['Paths']['logfile']

        # Take two reads for measurements to stabilise
        window.addstr( 0,0, "STARTING 1 ................" )
        window.refresh()
        arraytemp = acquire_temperatures( spi )    
        window.addstr( 0,0, "STARTING 2 ..................." )
        window.refresh()
        arraytemp = acquire_temperatures( spi )    

        currentepochtime = time.time()
        nextacqtime = currentepochtime + display_interval
        nextlogtime = currentepochtime + logging_interval
        nextacqflag = False
        nextlogflag = False

        # Main Loop
        while True:
            currentepochtime = time.time()
            nextacq = str( round( abs( nextacqtime - currentepochtime ) , 0 ) )
            nextlog = str( round( abs( nextlogtime - currentepochtime ) , 0 ) )

            if ( nextacqflag == False ):
                nextacqtime = currentepochtime + display_interval
                nextacqflag = True
            else:
                if ( currentepochtime > nextacqtime ):
                    nextacq = "Aquiring"
                    nextlog = "Hold"
                    output_to_screen( window, arraytemp, nextacq, nextlog )
                    arraytemp = acquire_temperatures( spi )    
                    nextacqflag = False

            if ( logoutput == 'ON' ):
                if ( nextlogflag == False ):
                    nextlogtime = currentepochtime + logging_interval
                    nextlogflag =  True
                else:
                    if ( currentepochtime > nextlogtime ):
                        nextacq = "Hold"
                        nextlog = "Logging"
                        output_to_screen( window, arraytemp, nextacq, nextlog )
                        arraytemp = acquire_temperatures( spi )    
                        output_to_log( window, arraytemp, logfile, logfile_maximum, currentepochtime )
                        nextlogflag = False
            else:
                nextlog = "Logging Disabled"

            output_to_screen( window, arraytemp, nextacq, nextlog )

            ch = window.getch()
            if ( (ch == ord('x')) or  (ch == ord('X')) ):
                curses.endwin()
                break

            time.sleep(1)

###-----------------------------------------------------------------------------###

# Main Line Code

# Check only one copy of app running.
runcount = 0
for pid in psutil.pids():
    p = psutil.Process(pid)
    taskname = p.name()
    if ( taskname == "TCRead.py" ):
        runcount = runcount + 1

if ( runcount > 1 ):    
    print( "This Program is already running. Only one instance of the program can run at once." )
    print( "**********************************************************************************" )
    exit( 1 )

os.system('clear')
curses.wrapper( mainlinecode )
